#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pywisp_emibcn.aircontrol import ACSession
from pprint import pprint

class Wisp():
   '''WISP interface to define infrastructure and host matching and authentication mechanisms'''
   
   ac_conf = {
      "url": "https://localhost:9082",
      "user": "admin",
      "password": "admin",
   }
   __ac = None
   __log = None
   __getlogger = None

   def __init__(self, ac_conf=None, getlog=None):
   
      if getlog:
         self.getlogger = getlog
         
      if ac_conf:
         self.ac_conf = ac_conf
         self.log.debug("__init__: Loaded AC configuration: %s" % (self.ac_conf) )

   @property
   def log(self):
      if not self.__log:
         self.__log = self.getlogger(__name__)

      return self.__log

   @log.setter
   def log(self, log):
      self.__log = log

   @property
   def getlogger(self):
      if not self.__getlogger:
         noop = lambda x, y: x
         return lambda name: type('fakelogger', (object,), {'error': print, 'warning': print, 'info': noop, 'debug': noop})()

      return self.__getlogger

   @getlogger.setter
   def getlogger(self, getlogger):
      self.__getlogger = getlogger
      self.log = None


   @property 
   def ac(self):
   
      if not self.__ac:
         self.__ac = ACSession(self.ac_conf['url'], self.ac_conf['user'], self.ac_conf['password'])
         self.__ac.login()
      
      return self.__ac

   @ac.setter
   def ac(self, ac):
      self.__ac = ac

      
   def get_host(self, name, deep=False, from_br=None):
      raise NotImplementedError("Should have implemented `get_host` method")


   def get_ac_devices(self):
      '''Generate antennas list (with credentials)'''
      raise NotImplementedError("Should have implemented `get_ac_devices` method")

   
   def get_ac_brs(self):
      '''Generate antennas list (with credentials) only of BRs'''
      raise NotImplementedError("Should have implemented `get_ac_brs` method")

   
   def get_mt_devices(self):
      '''Generate antennas list (with credentials)'''
      raise NotImplementedError("Should have implemented `get_mt_devices` method")


   def get_aircontrol_deep(self, name, from_br=None):
      self.log.info("Download BRs...")
      
      repetidors = self.get_ac_brs(from_br=None)
      
      self.log.info("BRs found: %s" % (len(repetidors)))

      clients_total = []
      clients_wifi = []
      for repetidor in repetidors:
         try:
            self.log.info("Download wifi stations from %s" % (repe.name))
            clients_wifi = repetidor.getWifiStations()
         except (KeyboardInterrupt, SystemExit):
            raise
         except Exception as e:
            self.log.warning("There was a problem connecting to %s: %s" % (repe.name, str(e)))
            continue
         
         clients = [
            {
               'deviceId': -1,
               'properties': {
                  'hostname': cw['remote']['hostname'],
                  'ip': cw['lastip'],
                  'mac': cw['mac'],
               }
            }
            for cw in clients_wifi
               if not name or \
                  (
                     name in cw['mac'].lower() or \
                     name in cw['lastip'] or \
                     ('remote' in cw and name in cw['remote']['hostname'].lower())
                  )
         ]
         
         for client in clients:
            self.log.info("%s - %s - %s" % (client['properties']['hostname'], client['properties']['mac'], client['properties']['ip']))
         
         clients_total += clients
      
      
      return clients_total


   
   def ac_reorder_branches(self):
      # Get devices list
      devices = self.get_ac_devices( self.ac.getDevices() )
      
      # Separate BRs and the rest of clients
      brs = []
      clients = []
      for antena in antenas:
         if antena.data['properties']['wlanOpModeString'] != 'sta':
            brs.append(antena)
         else:
            clients.append(antena)

      # For each BR
      done = 0
      patchList = []
      for br in brs:
         self.log.info("- %s (%s)" % (br.name, br.id))
         # For each client connectetd to it's SSID
         for client in clients:
            if client.ssid == br.ssid:
               if client.branch == br.id:
                  self.log.info("   - %s (%s)" % (client.name, client.ssid))
               else:
                  done += 1
                  self.log.info("   - Move {client} ({clid} - {clessid}) from {branch} to {brname} ({brid})".format(
                     client=client.name, 
                     clid=client.id, 
                     clessid=client.ssid, 
                     branch=client.branch, 
                     brname=br.name, 
                     brid=br.id))
                  
                  patchList.append(
                     self.ac.patchDeviceCreate(client.id, {"parentId": br.id})
                  )

      # Send patches
      self.ac.patchDeviceList(patchList)

      self.log.info("Done: %d" % (done + 1))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pywisp_emibcn.sshdevice import SSHDevice

STATUS = {
   "bound": "online",
   "waiting": "offline",
}

class MTDevice(SSHDevice):
   '''Mikrotik device class'''
   
   product = ""
   data    = {}
   
   
   def __init__(self, data, *args, **kwargs):
      '''Set specific Mikrotik values'''
      
      if 'host-name' in data:
         kwargs['name'] = data['host-name']
      if 'address' in data:
         kwargs['ip'] = data['address']
      if 'mac-address' in data:
         kwargs['mac'] = data['mac-address']
      if 'status' in data:
         kwargs['status'] = STATUS[data['status']]
      
      # Save original data
      self.data = data
      
      super().__init__(*args, **kwargs)
      
   
   def backup(self, path):
      '''Backup Mikrotik device config'''
      self.login()
      
      # Use MT export tool
      stdin, stdout, stderr = self.command("/export")
      
      # Save stdout to backup file
      with open(path + '/' + self.backup_file, "wb") as myfile:
         myfile.write(stdout.read())
   
   
   def parse_list(self, stdout):
      '''Parses a tipical 'terse' Mikrotik list'''
      
      # Consume first line (column titles, flags)
      lines = [""]
      index = 0
      
      # Put every item info into a line
      for line in stdout.readlines():
         # Primera línia d'un element
         if line.strip() == "":
            lines.append("")
            index = len(lines) - 1
            continue
         
         lines[index] += line.strip() + " "
      
      # Remove last empty line
      del lines[index]
      
      # Parse each line into a dict
      list = []
      for line in lines:
         variables = line.split()
         
         element = {}
         for variable in variables:
            if '=' in variable:
               var_val = variable.split('=', 1)
               element[var_val[0]] = var_val[1].strip('"')
            elif variable.isdigit():
               element['index'] = variable
         
         if len(element) > 0:
            list.append(element)
      
      del lines
      
      return list
   
   
   def parse_values(self, stdout):
      '''Parses a tipical 'variable: value' list'''
      
      element = {}
      for line in stdout:
         
         var_val = [x.strip() for x in line.strip().split(":", 1)]
         
         if len(var_val) < 2:
            continue
         
         while var_val[1][-1] == ',':
            var_val[1] += stdout.readline().strip()
         
         element[ var_val[0].strip() ] = var_val[1].strip()
      
      return element
   
   
   def getName(self):
      command = ':global idt [/system identity get name]; :put $idt;'
      stdin, stdout, stderr = self.command(command)
      
      self.name = stdout.readline().strip()
   
   
   def getDHCPLeases(self, name="", bound=True):
      
      command = '/ip dhcp-server lease print detail without-paging'
      where = ""
      
      # Filter by host-name with a regexp
      if name != "":
         where += ' host-name~"{}"'.format(name)
      
      # Filter (un)bound devices
      if bound != None:
         where += ' status{}=bound'.format('' if bound else '!')
      
      # Add filters to command
      if where != "":
         command += ' where' + where
      
      #stdin, stdout, stderr = self.command(":put \"" + command + "\"")
      stdin, stdout, stderr = self.command(command)
      
      return self.parse_list(stdout)
   

   def getWifiStatus(self):
      command = '/interface wireless registration-table print without-paging'
      command = "/interface wireless monitor 0 once"
      stdin, stdout, stderr = self.command(command)
      
      
      return self.parse_values(stdout)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import argparse, configparser, os, pkgutil, logging
from importlib import import_module
from pprint import pprint

# Internal imports
from pywisp_emibcn.wisp import Wisp
from pywisp_emibcn.sshdevice import backup_devices

class PyWisp():

    log = None
    args = None
    arg_parser = None
    args_sub_parser = None
    config = None
    wisp = None

    # Permet mostrar l'epilog amb la llista ben formatada, mentre es
    # mostren els arguments i els seus defaults formatats correctament
    # https://stackoverflow.com/questions/18462610/argumentparser-epilog-and-description-formatting-in-conjunction-with-argumentdef
    class MyCustomFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
        pass


    def __init__(self, wisp=None):
        '''Parses arguments and configurations'''

        self.log = self.setup_logger()

        self.args, self.arg_parser, self.arg_sub_parser = self.parse_arguments()
        self.config, self.wisp = self.parse_configuration(self.args.conf)

        # Allow to create a WISP externally (useful when imported as a library)
        if wisp:
            self.wisp = wisp


    def setup_logger(self, name=__name__):
        '''Setup a logger'''
        class OneLineExceptionFormatter(logging.Formatter):
             def formatException(self, exc_info):
                  result = super().formatException(exc_info)
                  return repr(result)

             def format(self, record):
                  result = super().format(record)
                  if record.exc_text:
                        result = result.replace("n", "")
                  return result

        handler = logging.StreamHandler()
        formatter = OneLineExceptionFormatter(logging.BASIC_FORMAT)
        handler.setFormatter(formatter)
        log = logging.getLogger(name)
        log.setLevel(os.environ.get("LOGLEVEL", "INFO"))
        log.addHandler(handler)

        return log


    def parse_device(self, device):
        '''Parses a device using arguments passed to program'''

        if 'getname' in self.args and self.args.getname:
            print(device.name)
        if 'getid' in self.args and self.args.getid:
            print(device.id)
        if 'getip' in self.args and self.args.getip:
            print(device.ip)
        if 'getmac' in self.args and self.args.getmac:
            print(device.mac)
        if 'getstatus' in self.args and self.args.getstatus:
            print(device.status)
        if 'getdhcp' in self.args and self.args.getdhcp:
            pprint(device.getDHCPLeases(bound=None))
        if 'getwifi' in self.args and self.args.getwifi:
            pprint(device.getWifiStatus())
        if 'getwifistations' in self.args and self.args.getwifistations:
            pprint(device.getWifiStations())
        if 'getjson' in self.args and self.args.getjson:
            pprint(device.data)
        if 'url' in self.args and self.args.url:
            print(self.wisp.ac.getDevicesURL([device.id])[0]['url'])
        if 'cmd' in self.args and self.args.cmd:
            stdin, stdout, stderr = device.command(args.cmd)
            print(stdout.read().decode(), end="")
            print(stderr.read().decode(), end="")
        if 'ssh' in self.args and self.args.ssh:
            device.shell()


    def parse_arguments(self, parser=argparse.ArgumentParser(formatter_class=MyCustomFormatter)):
        '''Parses arguments passed to program into a dict'''

        parser.add_argument("--conf", type=str,
                                  default="{}/{}".format(os.getenv('HOME'), '.pywisp'),
                                  help="Reads configuration from this file instead of default")

        sp = parser.add_subparsers()

        b_ac = sp.add_parser("backup_ac",formatter_class=self.MyCustomFormatter,
                                  help="Backup all AirControl devices")
        b_ac.add_argument("backup_ac_path", type=str, nargs='?',metavar="PATH",
                                  help="Directory in which save backup files")
        b_ac.add_argument("--retries",
                                  action="store_true", default=3,
                                  help="Retries for every device before stop trying")

        b_mt = sp.add_parser("backup_mt",formatter_class=self.MyCustomFormatter,
                                  help="Backup all Mikrotik devices")
        b_mt.add_argument("backup_mt_path", type=str, nargs='?',metavar="PATH",
                                  help="Directory in which save backup files")
        b_mt.add_argument("--retries",
                                  action="store_true", default=3,
                                  help="Retries for every device before stop trying")

        reorder = sp.add_parser("reorder_ac",formatter_class=self.MyCustomFormatter,
                                  help="Reorder branches from AirControl devices")

        host_parser = sp.add_parser("host", formatter_class=self.MyCustomFormatter,
                                  help="Find device by it's hostname, MAC or IP")
        host_parser.add_argument("host", type=str,
                                  help="Devices hostname, MAC or IP")

        host_parser.add_argument("--deep",
                                  action="store_true",
                                  help="Find device by it's hostname, MAC or IP, using all BRs station list as haystack")
        host_parser.add_argument("--from-br", type=str,
                                  help="Deep find only in this BR")

        host_parser.add_argument("--getname",
                                  action="store_true",
                                  help="Gets device name")
        host_parser.add_argument("--getjson",
                                  action="store_true",
                                  help="Gets device full data")
        host_parser.add_argument("--getip",
                                  action="store_true",
                                  help="Gets device IP")
        host_parser.add_argument("--getid",
                                  action="store_true",
                                  help="Gets device ID")
        host_parser.add_argument("--getmac",
                                  action="store_true",
                                  help="Gets device MAC")
        host_parser.add_argument("--getdhcp",
                                  action="store_true",
                                  help="Gets device's DHCP leases/stations")
        host_parser.add_argument("--getwifi",
                                  action="store_true",
                                  help="Gets device's wifi status")
        host_parser.add_argument("--getwifistations",
                                  action="store_true",
                                  help="Gets device's wifi stations")
        host_parser.add_argument("--getstatus",
                                  action="store_true",
                                  help="Gets device status")
        host_parser.add_argument("--url",
                                  action="store_true",
                                  help="Gets an authenticated URL to connect to the device using browser")

        host_parser.add_argument("--ssh",
                                  action="store_true",
                                  help="Connects to device using SSH")
        host_parser.add_argument("--cmd", type=str,
                                  help="Connects to device and run a command")

        args = parser.parse_args()

        self.log.debug('Arguments: %s' % args)

        return args, parser, sp


    def parse_configuration(self, file_name):
        '''Parses configuration file into a dict/class'''
        config = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())

        wisp = None
        wisp_conf = {'getlog': self.setup_logger}

        if os.path.isfile(file_name):

            # Populate the [env] section
            env_dict = {name: os.path.expandvars(value.replace('$', '$$')) for name, value in os.environ.items()}
            config.read_dict(
                {"env": env_dict},
                source='<env>')

            self.log.debug("Loading %s" % file_name)
            config.read(file_name)

            wisp_conf['ac_conf'] = None
            if 'ac' in config:
                wisp_conf['ac_conf'] = {name: value for name, value in config.items('ac')}
                self.log.debug("Loaded AC configuration:  %s" % wisp_conf['ac_conf'])

            if 'wisp' in config:
                self.log.debug("Detected extra class for managing wisp.")

                if 'path' in config['wisp'] and 'module' in config['wisp'] and 'class' in config['wisp']:
                    self.log.debug("Detected WISP: %s || %s || %s" % (config['wisp']['path'], config['wisp']['module'], config['wisp']['class']))
                    path = os.path.dirname( config['wisp']['path'] )

                    for (finder, name, _) in pkgutil.iter_modules([path]):
                        if name == config['wisp']['module']:
                            self.log.debug("Loading module %s from file '%s'..." % (config['wisp']['module'], config['wisp']['path']))
                            loader = finder.find_module(name)
                            mod = loader.load_module()
                            cls = getattr(mod, config['wisp']['class'])

                            self.log.debug("Merging class %s to our WISP definition..." % config['wisp']['class'])
                            wisp = type('Wisp' + config['wisp']['class'], (cls,Wisp), {})(**wisp_conf)
                            self.log.debug("Wisp Class: %s" % type(wisp).__name__)
            else:
                wisp = Wisp(**wisp_conf)

        else:
            self.log.warning("No config file")

        return config, wisp


def main():

    pywisp = PyWisp()

    # Backup everything!
    if 'backup_ac_path' in pywisp.args:
        path = pywisp.args.backup_ac_path
        if not path and 'backup' in pywisp.config and 'ac' in pywisp.config['backup']:
            path = pywisp.config['backup']['ac']

        retries = pywisp.args.retries
        if not retries and 'backup' in pywisp.config and 'retries' in pywisp.config['backup']:
            retries = pywisp.config['backup']['retries']

        pywisp.log.debug('Backup AC devices to %s' % (path))
        backup_devices(pywisp.wisp.get_ac_devices(), path, retries=retries)

    elif 'backup_mt_path' in pywisp.args:
        path = pywisp.args.backup_mt_path
        if not path and 'backup' in pywisp.config and 'mt' in pywisp.config['backup']:
            path = pywisp.config['backup']['mt']

        retries = pywisp.args.retries
        if not retries and 'backup' in pywisp.config and 'retries' in pywisp.config['backup']:
            retries = pywisp.config['backup']['retries']

        pywisp.log.debug('Backup MT devices to %s' % (path))
        backup_devices(pywisp.wisp.get_mt_devices(), path, retries=retries)

    # Reorder AirControl branches
    elif 'reorder_ac' in pywisp.args:
        pywisp.log.debug('Reorder branches!')
        pywisp.wisp.ac_reorder_branches()

    # Find host and print info about it or perform actions on it
    elif 'host' in pywisp.args:
        # Optional lower host (insensitive)
        pywisp.args.host = pywisp.args.host.lower().strip()

        # Get devices
        devices = pywisp.wisp.get_host(pywisp.args.host, deep=pywisp.args.deep, from_br=pywisp.args.from_br)
        if devices == None or len(devices) == 0:
            pywisp.log.error("No devices found: '%s'" % (pywisp.args.host))
            return 1

        # If its not a list, convert into it
        if type(devices) is not list:
            devices = [ devices ]

        # Apply actions for every device found
        for device in devices:
            pywisp.parse_device(device)

    return 0


if __name__ == "__main__":
    exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime, paramiko, base64, os, socket
from termcolor import colored
from pprint import pprint
try:
    import interactive
except ImportError:
    from . import interactive


class SSHDevice:

    ip         = ""
    mac        = ""
    __name    = ""
    username = ""
    password = ""
    rsa        = ""
    status    = ""
    backup_file_base = ""

    client = False

    def __init__(self, ip="", mac="", name="", username="", password="", rsa="", status="", backup_file=""):

        # Set minimal device data
        self.ip         = ip
        self.mac        = mac
        self.name      = name
        self.username = username
        self.password = password
        self.rsa        = rsa
        self.status    = status
        self.backup_file_base = backup_file

    # Use setters anf getters for properties needing expensive actions
    @property
    def name(self):
        if not self.__name or self.__name == "":
            self.getName()
        return self.__name

    @name.setter
    def name(self, name):
        self.__name = name
        if name and name != "":
            self.setBackupName()

    @property
    def backup_file(self):
        if not self.__backup_file or self.__backup_file == "":
            self.setBackupName()
        return self.__backup_file

    @backup_file.setter
    def backup_file(self, backup_file):
        self.__backup_file = backup_file

    def getName(self):
        raise NotImplementedError("Should have implemented `getName` method")

    def setBackupName(self, backup_file=""):
        '''Set backup filename'''

        if backup_file:
            self.backup_file_base = backup_file

        # Set backup file name
        if not self.backup_file_base:
            today = datetime.date.today()
            self.backup_file = "{name}.{date}.bkp".format( name=self.name, date=today.strftime('%Y-%m-%d') )
        else:
            self.backup_file = self.backup_file_base


    def login(self):
        '''Open SSH connection only if it is not already opened'''
        if self.client == False:

            # Start SSH connection
            self.client = paramiko.SSHClient()
            self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            # Try login with user/password
            try:
                self.client.connect(self.ip, username=self.username, password=self.password, timeout=5, allow_agent=False, look_for_keys=False)
            except:
                # Try login with RSA key
                key = paramiko.RSAKey.from_private_key_file(self.rsa)
                self.client.connect(self.ip, username=self.username, timeout=5, pkey=key)


    def logout(self):
        '''Close SSH connection only if it is opened'''
        if self.client != False:
            self.client.close()
            # Higiene
            del self.client
            self.client = False


    def command(self, command):
        '''Send command to device and return (stdin, stdout, stderr) streams tuple'''
        self.login()

        return self.client.exec_command(command, timeout=5)


    def shell(self, *args, **kwargs):
        '''Opens a TTY shell'''
        self.login()

        if 'self' in kwargs:
            del kwargs['self']

        chan = self.client.invoke_shell(*args, **kwargs)
        interactive.interactive_shell(chan)
        chan.close()


    # Higiene
    def __del__(self):
        self.logout()

    def __str__(self):
        if self.mac:
            return u"{} : {} - {}".format(self.name, self.mac, self.ip)
        else:
            return u"{} : {}".format(self.name, self.ip)



def backup_devices_list(devices, path):
    '''Do backup on an ACDevice list'''
    i = 1
    failed = []

    for device in devices:

        # Debug:
        #if i == 10:
        #    break

        print(u"{index}.- {device}".format(index=i, device=str(device)))

        # Debug:
        #if i == 1:
        #    pprint(device.data)
        #    break

        warning = ""
        file = path + "/" + device.backup_file
        if os.path.exists(file) and os.stat(file).st_size > 0:
            print(u"    " + colored("[WARNING] Backup ja realitzat. Saltem.", 'yellow', attrs=['bold']))
            continue

        try:
            device.backup(path)
        except paramiko.ssh_exception.AuthenticationException as e:
            warning = u"[WARNING] Credencials incorrectes! (" + str(e) + ")"
        except paramiko.ssh_exception.NoValidConnectionsError as e:
            warning = u"[WARNING] No es pot establir connexió al port 22! (" + str(e) + ")"
        except socket.timeout as e:
            warning = u"[WARNING] Servidor no abastable! (" + str(e) + ")"
        except KeyboardInterrupt as e:
            raise e
        except Exception as e:
            warning = u"[WARNING] Excepció no gestionada: " + str(e)
        finally:
            device.logout()

        if warning != "":
            device.warning = warning
            print(u"    " + colored(warning, 'red', attrs=['bold']))
            failed.append(device)

        # Higiene
        else:
            del device

        i += 1

    return failed


def backup_devices(devices, path, retries=3):
    # Ensure backup dir exists
    print(u"Make dir: " + path)
    os.makedirs(path, exist_ok=True)

    failed = devices
    ok = 0
    while retries > 0:

        total = len(failed)

        # Do backup and get failed list
        failed = backup_devices_list(failed, path)

        # Sum non-failed to 'ok' counter
        ok += total - len(failed)

        # If list is empty, break while
        if len(failed) == 0:
            break

        # Decrease counter
        retries -= 1
        if retries > 0:
            print(colored(u"\nTornem a intentar amb les antenes que hagin fallat (queden {} intents, {} fallats)\n".format(retries-1, len(failed)), 'white', attrs=['bold']))

    # Print totals
    print(u"\n")
    print(colored(u"Descarregats {ok} backups".format(ok=ok), 'green', attrs=['bold']))
    print(colored(u"NO Descarregats {ko} backups".format(ko=len(failed)), 'red', attrs=['bold']))
    print(u"\n")

    # Print failed
    if len(failed) > 0:
        print(u"Failed %d devices:" % (len(failed)))
        for f in failed:
            print(u" - {device}: {warning}".format(device=str(f), warning=colored(f.warning, 'red', attrs=['bold'])))

        print(u"\n")




from setuptools import setup, find_packages

VERSION = (0, 0, 8)
__version__ = '.'.join(map(str, VERSION))

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read().replace('(/', '(https://github.com/emibcn/pywisp/blob/master/')

setup(
    name = 'pywisp_emibcn',
    version = __version__,
    description = "Access AirControl2, AirOS & Mikrotik",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author = "https://github.com/emibcn",
    author_email = "emibcn.github@gmail.com",
    url = "https://github.com/emibcn/pywisp",
    project_urls={
        "Bug Tracker": "https://github.com/emibcn/pywisp/issues",
    },
    license = "GPLv3",
    platforms = ["any"],
    packages = ['pywisp_emibcn'],
    package_dir = {"": "."},
    py_modules=['pywisp', 'pywisp/mikrotik', 'pywisp/aircontrol', 'pywisp/sshdevice'],
    scripts = ["bin/pywisp"],
    install_requires = ['requests', 'paramiko', 'termcolor', "argparse", "configparser"],
    include_package_data = True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Environment :: Other Environment",
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, requests, ipaddress, datetime, json
from pywisp_emibcn.sshdevice import SSHDevice

from pprint import pformat

# Disable warnings
import urllib3
urllib3.disable_warnings()

# Internal utils
def ip2int(addr):
   return int(ipaddress.IPv4Address(addr))


def int2ip(addr):
   return str(ipaddress.IPv4Address(addr))


def is_ip(addr):
   try:
      ipaddress.ip_address(addr)
   except:
      return False
   
   return True


def get_client_from_wifi_station(client_wifi):
   return {
            'deviceId': -1,
            'properties': {
               'hostname': client_wifi['remote']['hostname'] if 'remote' in client_wifi else client_wifi['name'],
               'ip': client_wifi['lastip'],
               'mac': client_wifi['mac'],
            }
         }


STATUS = ["unknown", "not monitored", "online", "offline"]

class ACDevice(SSHDevice):
   '''AirControl/Ubiquiti device class'''
   
   id      = 0
   ssid    = ""
   branch  = ""
   product = ""
   data    = {}
   
   
   def __init__(self, json, ac=None, *args, **kwargs):
      '''Set specific AirOS values'''
      
      today = datetime.date.today()
      
      if ac and 'hostname' in json['properties']:
         d = ac.getDevices(name=json['properties']['hostname'])
         if len(d) > 0:
            json = d[0]
      
      self.data = json
         
      #pprint(json)
      
      # Assign data from original device JSON
      self.id = json['deviceId']
      
      if 'hostname' in json['properties']:
         kwargs['name']        = json['properties']['hostname']
      if 'mac' in json['properties']:
         kwargs['mac'] = json['properties']['mac']
      if 'ip' in json['properties']:
         kwargs['ip'] = int2ip(json['properties']['ip'])
      if 'status' in json['properties']:
         if json['properties']['status'] in STATUS:
            kwargs['status'] = STATUS[ json['properties']['status'] ]
         else:
            kwargs['status'] = json['properties']['status']
      
      # Set specific AirOS properties
      if 'parentId' in json:
         self.branch = json['parentId']
      if 'essid' in json['properties']:
         self.ssid    = json['properties']['essid']
      if 'product' in json['properties']:
         self.product = json['properties']['product']
      
      super().__init__(*args, **kwargs)

   
   def getName(self):
      '''Get device name from the device itself'''
      
      command = 'uname -n'
      stdin, stdout, stderr = self.command(command)
      
      self.name = stdout.readline().strip()
   
   
   def setBackupName(self, backup_file=""):
      '''Append '.tar' to backup filename'''
      
      super().setBackupName(backup_file=backup_file)
      self.backup_file += ".tar"
   
   
   def backup(self, path):
      
      self.login()
      
      # No sFTP server on Ubiquiti systems. Let's do a tar.
      # Send tar command with stdout as local backup file
      files = ["/tmp/system.cfg", "/etc/persistent/rc.prestart", "/etc/persistent/rc.poststart"]
      command = 'tar -c -f - {}'.format(" ".join('"{}"'.format(f) for f in files))
      stdin, stdout, stderr = self.command(command)
      
      # Write tar command output to local tar file
      with open(path + '/' + self.backup_file, "wb") as myfile:
         myfile.write(stdout.read())
      

   def getWifiStatus(self):
      status = {}
      
      for var, value in self.data['properties'].items():
         if var.startswith('chain') or \
               var.startswith('channel') or \
               var.startswith('freq') or \
               var.startswith('essid') or \
               var.startswith('noise') or \
               var.startswith('distance') or \
               var.startswith('signal'):
            status[var] = value
      
      return status
   
   
   def getWifiStations(self):
      command = "wstalist ath0"
      stdin, stdout, stderr = self.command(command)
      
      return json.loads(stdout.read().decode())
   
   
   def __str__(self):
      return u"{} : {} - {} ({})".format(self.name, self.mac, self.ip, self.id)

   
class ACSession():
   cookies = {}
   URL_path = "/api/v1"
   URL = ""
   username = ""
   password = ""
   devices = None
   
   def __init__(self, URL, username, password):
      '''Assign login parameters'''
      self.URL = URL
      self.username = username
      self.password = password
         
   
   def login(self):
      '''Login to AirControl API server'''
      loginData = {
         'username': self.username,
         'password': self.password,
         'eulaAccepted':True,
      }
      
      # Login to server
      resp = requests.post(
         self.URL + self.URL_path + '/login', 
         json=loginData,
         headers={
            'Accept':'application/json', 
            'Content-Type':'application/json'
         },
         verify=False)
      
      # This means something went wrong.
      if resp.status_code != 200:
         raise Exception(u'GET /login/ {}: {}'.format(resp.status_code, resp.text))
      
      # Save cookies (session)
      self.cookies = resp.cookies
   
   
   def sendRequest(self, path, method="get", body={}):
      '''Send request to AirControl API server using cookies from login'''
      URL = self.URL + self.URL_path + path
      arguments = {
         'cookies':self.cookies,
         'headers': {
            'Accept':'application/json', 
            'Content-Type':'application/json'
         },
         'verify':False,
      }
      
      if method == 'get':
         resp = requests.get(URL, **arguments)
      elif method == 'post':
         resp = requests.post(URL, data=str(body), **arguments)
      elif method == 'patch':
         resp = requests.patch(URL, data=str(body), **arguments)
      
      # This means something went wrong.
      if resp.status_code > 299:
         raise Exception(u'{} {} {}: {} ({})'.format(method.upper(), path, resp.status_code, resp.text, pformat(body)))
      
      return resp
      
      
   def getDevices(self, name_starts=None, name=None, ip=None, mac=None):
      '''Get devices list, as a list of dicts (from JSON data)'''
      
      if mac is not None:
         return self.getDeviceByMac(mac)
         
      if not self.devices:
         self.devices = self.sendRequest("/devices").json()['results']
      
      if name:
         name = name.lower()
      if name_starts:
         name_starts = name_starts.lower()
      if mac:
         mac = mac.lower()
      
      if not name_starts and not name and not ip and not mac:
         result = self.devices
      else:
         result = [
            dev
            for dev in self.devices
               if ( 'properties' in dev and 'hostname' in dev['properties'] ) and \
                  (
                     (name_starts and dev['properties']['hostname'].lower().startswith(name_starts) ) or \
                     (name        and name in dev['properties']['hostname'].lower()                 ) or \
                     (mac         and mac  in dev['properties']['mac'].lower()                      ) or \
                     (ip          and ip   in int2ip(dev['properties']['ip'])                       )    \
                  ) 
         ]

      return result
   
   
   def getDeviceByMac(self, mac):
      '''Gets device by it's MAC address'''
      return self.sendRequest("/devices/mac/{}".format(mac)).json()
   
   
   def getDeviceStatus(self, id):
      '''Get device's URL'''
      req = {
         "metricSetId": 32,
         "from": 0,
         "to": 10,
         "scale": "seconds"
      }
      result = self.sendRequest("/devices/{}/metrics".format(id), method='post', body=json.dumps(req))
      return result.json()
   
   
   def getDevicesURL(self, list):
      '''Get device's URL'''
      result = self.sendRequest("/devices/webui", method='post', body=str(list))
      return result.json()['results']
   
   
   def patchDeviceCreate(self, deviceId, patchDevice):
      '''Creates single device basic properties patch'''
      
      acceptedMods = {
          "type": str,
          "ip": str,
          "domainName": str,
          "sshUserName": str,
          "sshPassword": str,
          "sshPort": int,
          "rememberSshSettings": bool,
          "httpPort": int,
          "useHttps": bool,
          "description": str,
          "tag": str,
          "parentId": int,
          "uplinkType": int,
          "overriddenServerAddress": {
            "ip": str,
            "port": int
          },
          "membershipTypeGateway": bool,
          "runDiscovery": bool,
          "lockPositionInTopology": bool,
          "lockIpAddress": bool,
          "connectedDirectly": bool,
          "sessionId": int,
          "deviceId": int
      }
      
      patch = {}
      for name,value in acceptedMods.items():
         if name in patchDevice and value == type(patchDevice[name]):
            patch[name] = patchDevice[name]
      
      patch['deviceId'] = deviceId
      
      return patch


   def patchDeviceList(self, patchList):
      '''Patch devices list basic properties'''
      
      return self.sendRequest(
                         "/devices/basic-properties", 
                         method='patch', 
                         body=json.dumps( patchList ))


   def patchDevice(self, deviceId, patchDevice):
      '''Patch device basic properties'''
      
      return self.patchDeviceList( [self.patchDeviceCreate(deviceId, patchDevice)] )
   
   
